"""
Key differences with COM UDFs:
* respects ints (COM always returns floats)
* returns 0 for empty cells. To get None like in COM, you need to set the formula to: =""
* caller range object not supported (caller address would easy to get though)
* reading datetime must be explicitly converted via dt.date / dt.datetime or parse_dates (pandas)
* writing datetime is now automatically formatting it as date in Excel
* categories aren't supported: replaced by namespaces
"""
import datetime as dt
import sys
from datetime import date, datetime

from nose.tools import assert_dict_equal

import xlwings as xw
from xlwings.pro import arg, func, ret

try:
    import numpy as np
    from numpy.testing import assert_array_equal

    def nparray_equal(a, b):
        try:
            assert_array_equal(a, b)
        except AssertionError:
            return False
        return True

except ImportError:
    np = None
try:
    import pandas as pd
    from pandas.testing import assert_frame_equal, assert_series_equal

    def frame_equal(a, b):
        try:
            assert_frame_equal(a, b)
        except AssertionError:
            return False
        return True

    def series_equal(a, b):
        try:
            assert_series_equal(a, b)
        except AssertionError:
            return False
        return True

except ImportError:
    pd = None


def dict_equal(a, b):
    try:
        assert_dict_equal(a, b)
    except AssertionError:
        return False
    return True


# Defaults
@func
def read_float(x):
    return x == 2


@func
def write_float():
    return 2


@func
def read_string(x):
    return x == "xlwings"


@func
def write_string():
    return "xlwings"


@func
def read_empty(x):
    return x is None


@func
@arg("x", dt.datetime)
def read_date(x):
    print(x)
    return x == datetime(2015, 1, 15)


@func
def write_date():
    return datetime(1969, 12, 31)


@func
@arg("x", dt.datetime)
def read_datetime(x):
    return x == datetime(1976, 2, 15, 13, 6, 22)


@func
def write_datetime():
    return datetime(1976, 2, 15, 13, 6, 23)


@func
def read_horizontal_list(x):
    return x == [1, 2]


@func
def write_horizontal_list():
    return [1, 2]


@func
def read_vertical_list(x):
    return x == [1, 2]


@func
def write_vertical_list():
    return [[1], [2]]


@func
def read_2dlist(x):
    return x == [[1, 2], [3, 4]]


@func
def write_2dlist():
    return [[1, 2], [3, 4]]


# Keyword args on default converters


@func
@arg("x", ndim=1)
def read_ndim1(x):
    return x == [2]


@func
@arg("x", ndim=2)
def read_ndim2(x):
    return x == [[2]]


@func
@arg("x", transpose=True)
def read_transpose(x):
    return x == [[1, 3], [2, 4]]


@func
@ret(transpose=True)
def write_transpose():
    return [[1, 2], [3, 4]]


@func
def read_dates_as1(x):
    x[0][1] = xw.to_datetime(x[0][1]).date()
    x[1][0] = xw.to_datetime(x[1][0]).date()
    return x == [[1, date(2015, 1, 13)], [date(2000, 12, 1), 4]]


@func
@arg("x", dt.date)
def read_dates_as2(x):
    return x == date(2005, 1, 15)


@func
def read_dates_as3(x):
    x[0][1] = xw.to_datetime(x[0][1])
    x[1][0] = xw.to_datetime(x[1][0])
    return x == [[1, datetime(2015, 1, 13)], [datetime(2000, 12, 1), 4]]


@func
@arg("x", empty="empty")
def read_empty_as(x):
    return x == [[1, "empty"], ["empty", 4]]


if sys.version_info >= (2, 7):
    # assert_dict_equal isn't available on nose for PY 2.6

    # Dicts
    @func
    @arg("x", dict)
    def read_dict(x):
        return dict_equal(x, {"a": 1, "b": "c"})

    @func
    @arg("x", dict, transpose=True)
    def read_dict_transpose(x):
        return dict_equal(x, {1: "c", "a": "b"})


@func
def write_dict():
    return {"a": 1, "b": "c"}


# Numpy Array
if np:

    @func
    @arg("x", np.array)
    def read_scalar_nparray(x):
        return nparray_equal(x, np.array(1))

    @func
    @arg("x", np.array)
    def read_empty_nparray(x):
        return nparray_equal(x, np.array(np.nan))

    @func
    @arg("x", np.array)
    def read_horizontal_nparray(x):
        return nparray_equal(x, np.array([1, 2]))

    @func
    @arg("x", np.array)
    def read_vertical_nparray(x):
        return nparray_equal(x, np.array([1, 2]))

    @func
    @arg("x", dt.datetime)
    def read_date_nparray(x):
        x = np.array(x)
        return nparray_equal(x, np.array(datetime(2000, 12, 20)))

    # Keyword args on Numpy arrays

    @func
    @arg("x", np.array, ndim=1)
    def read_ndim1_nparray(x):
        return nparray_equal(x, np.array([2]))

    @func
    @arg("x", np.array, ndim=2)
    def read_ndim2_nparray(x):
        return nparray_equal(x, np.array([[2]]))

    @func
    @arg("x", np.array, transpose=True)
    def read_transpose_nparray(x):
        return nparray_equal(x, np.array([[1, 3], [2, 4]]))

    @func
    @ret(transpose=True)
    def write_transpose_nparray():
        return np.array([[1, 2], [3, 4]])

    @func
    @arg("x", dt.date)
    def read_dates_as_nparray(x):
        x = np.array(x)
        return nparray_equal(x, np.array(date(2000, 12, 20)))

    @func
    @arg("x", np.array, empty="empty")
    def read_empty_as_nparray(x):
        return nparray_equal(x, np.array("empty"))

    @func
    def write_np_scalar():
        return np.float64(2)


# Pandas Series

if pd:

    @func
    @arg("x", pd.Series, header=False, index=False)
    def read_series_noheader_noindex(x):
        return series_equal(x, pd.Series([1, 2]))

    @func
    @arg("x", pd.Series, header=False, index=True)
    def read_series_noheader_index(x):
        return series_equal(x, pd.Series([1, 2], index=[10, 20]))

    @func
    @arg("x", pd.Series, header=True, index=False)
    def read_series_header_noindex(x):
        return series_equal(x, pd.Series([1, 2], name="name"))

    @func
    @arg("x", pd.Series, header=True, index=True)
    def read_series_header_named_index(x):
        return series_equal(
            x,
            pd.Series([1, 2], name="name", index=pd.Index([10, 20], name="ix")),
        )

    @func
    @arg("x", pd.Series, header=True, index=True)
    def read_series_header_nameless_index(x):
        print(x)
        return series_equal(x, pd.Series([1, 2], name="name", index=[10, 20]))

    @func
    @arg("x", pd.Series, header=True, index=2)
    def read_series_header_nameless_2index(x):
        ix = pd.MultiIndex.from_arrays([["a", "a"], [10, 20]])
        return series_equal(x, pd.Series([1, 2], name="name", index=ix))

    @func
    @arg("x", pd.Series, header=True, index=2)
    def read_series_header_named_2index(x):
        ix = pd.MultiIndex.from_arrays([["a", "a"], [10, 20]], names=["ix1", "ix2"])
        return series_equal(x, pd.Series([1, 2], name="name", index=ix))

    @func
    @arg("x", pd.Series, header=False, index=2)
    def read_series_noheader_2index(x):
        ix = pd.MultiIndex.from_arrays([["a", "a"], [10, 20]])
        return series_equal(x, pd.Series([1, 2], index=ix))

    @func
    @ret(pd.Series, index=False)
    def write_series_noheader_noindex():
        return pd.Series([1, 2])

    @func
    @ret(pd.Series, index=True)
    def write_series_noheader_index():
        return pd.Series([1, 2], index=[10, 20])

    @func
    @ret(pd.Series, index=False)
    def write_series_header_noindex():
        return pd.Series([1, 2], name="name")

    @func
    def write_series_header_named_index():
        return pd.Series([1, 2], name="name", index=pd.Index([10, 20], name="ix"))

    @func
    @ret(pd.Series, index=True, header=True)
    def write_series_header_nameless_index():
        return pd.Series([1, 2], name="name", index=[10, 20])

    @func
    @ret(pd.Series, header=True, index=2)
    def write_series_header_nameless_2index():
        ix = pd.MultiIndex.from_arrays([["a", "a"], [10, 20]])
        return pd.Series([1, 2], name="name", index=ix)

    @func
    @ret(pd.Series, header=True, index=2)
    def write_series_header_named_2index():
        ix = pd.MultiIndex.from_arrays([["a", "a"], [10, 20]], names=["ix1", "ix2"])
        return pd.Series([1, 2], name="name", index=ix)

    @func
    @ret(pd.Series, header=False, index=2)
    def write_series_noheader_2index():
        ix = pd.MultiIndex.from_arrays([["a", "a"], [10, 20]])
        return pd.Series([1, 2], index=ix)

    @func
    @arg("x", pd.Series, parse_dates=True)
    def read_timeseries(x):
        return series_equal(
            x,
            pd.Series(
                [1.5, 2.5],
                name="ts",
                index=[datetime(2000, 12, 20), datetime(2000, 12, 21)],
            ),
        )

    @func
    @ret(pd.Series)
    def write_timeseries():
        return pd.Series(
            [1.5, 2.5],
            name="ts",
            index=[datetime(2000, 12, 20), datetime(2000, 12, 21)],
        )

    @func
    @ret(pd.Series, index=False)
    def write_series_nan():
        return pd.Series([1, np.nan, 3])


# Pandas DataFrame

if pd:

    @func
    @arg("x", pd.DataFrame, index=False, header=False)
    def read_df_0header_0index(x):
        return frame_equal(x, pd.DataFrame([[1, 2], [3, 4]]))

    @func
    @ret(pd.DataFrame, index=False, header=False)
    def write_df_0header_0index():
        return pd.DataFrame([[1, 2], [3, 4]])

    @func
    @arg("x", pd.DataFrame, index=False, header=True)
    def read_df_1header_0index(x):
        return frame_equal(x, pd.DataFrame([[1, 2], [3, 4]], columns=["a", "b"]))

    @func
    @ret(pd.DataFrame, index=False, header=True)
    def write_df_1header_0index():
        return pd.DataFrame([[1, 2], [3, 4]], columns=["a", "b"])

    @func
    @arg("x", pd.DataFrame, index=True, header=False)
    def read_df_0header_1index(x):
        return frame_equal(x, pd.DataFrame([[1, 2], [3, 4]], index=[10, 20]))

    @func
    @ret(pd.DataFrame, index=True, header=False)
    def write_df_0header_1index():
        return pd.DataFrame([[1, 2], [3, 4]], index=[10, 20])

    @func
    @arg("x", pd.DataFrame, index=2, header=False)
    def read_df_0header_2index(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            index=pd.MultiIndex.from_arrays([["a", "a", "b"], [1, 2, 1]]),
        )
        return frame_equal(x, df)

    @func
    @ret(pd.DataFrame, index=2, header=False)
    def write_df_0header_2index():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            index=pd.MultiIndex.from_arrays([["a", "a", "b"], [1, 2, 1]]),
        )
        return df

    @func
    @arg("x", pd.DataFrame, index=1, header=1)
    def read_df_1header_1namedindex(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[1, 2],
            columns=["c", "d", "c"],
        )
        df.index.name = "ix1"
        return frame_equal(x, df)

    @func
    def write_df_1header_1namedindex():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[1, 2],
            columns=["c", "d", "c"],
        )
        df.index.name = "ix1"
        return df

    @func
    @arg("x", pd.DataFrame, index=1, header=1)
    def read_df_1header_1unnamedindex(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[1, 2],
            columns=["c", "d", "c"],
        )
        return frame_equal(x, df)

    @func
    def write_df_1header_1unnamedindex():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[1, 2],
            columns=["c", "d", "c"],
        )
        return df

    @func
    @arg("x", pd.DataFrame, index=False, header=2)
    def read_df_2header_0index(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        return frame_equal(x, df)

    @func
    @ret(pd.DataFrame, index=False, header=2)
    def write_df_2header_0index():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        return df

    @func
    @arg("x", pd.DataFrame, index=1, header=2)
    def read_df_2header_1namedindex(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[1, 2],
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        df.index.name = "ix1"
        return frame_equal(x, df)

    @func
    def write_df_2header_1namedindex():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[1, 2],
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        df.index.name = "ix1"
        return df

    @func
    @arg("x", pd.DataFrame, index=1, header=2)
    def read_df_2header_1unnamedindex(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[1, 2],
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        return frame_equal(x, df)

    @func
    def write_df_2header_1unnamedindex():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[1, 2],
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        return df

    @func
    @arg("x", pd.DataFrame, index=2, header=2)
    def read_df_2header_2namedindex(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            index=pd.MultiIndex.from_arrays(
                [["a", "a", "b"], [1, 2, 1]], names=["x1", "x2"]
            ),
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        return frame_equal(x, df)

    @func
    def write_df_2header_2namedindex():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            index=pd.MultiIndex.from_arrays(
                [["a", "a", "b"], [1, 2, 1]], names=["x1", "x2"]
            ),
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        return df

    @func
    @arg("x", pd.DataFrame, index=2, header=2)
    def read_df_2header_2unnamedindex(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            index=pd.MultiIndex.from_arrays([["a", "a", "b"], [1, 2, 1]]),
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        return frame_equal(x, df)

    @func
    def write_df_2header_2unnamedindex():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            index=pd.MultiIndex.from_arrays([["a", "a", "b"], [1, 2, 1]]),
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        return df

    @func
    @arg("x", pd.DataFrame, index=2, header=1)
    def read_df_1header_2namedindex(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            index=pd.MultiIndex.from_arrays(
                [["a", "a", "b"], [1, 2, 1]], names=["x1", "x2"]
            ),
            columns=["a", "d", "c"],
        )
        return frame_equal(x, df)

    @func
    def write_df_1header_2namedindex():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            index=pd.MultiIndex.from_arrays(
                [["a", "a", "b"], [1, 2, 1]], names=["x1", "x2"]
            ),
            columns=["a", "d", "c"],
        )
        return df

    @func
    @arg("x", pd.DataFrame, parse_dates=True)
    def read_df_date_index(x):
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[datetime(1999, 12, 13), datetime(1999, 12, 14)],
            columns=["c", "d", "c"],
        )
        return frame_equal(x, df)

    @func
    def write_df_date_index():
        df = pd.DataFrame(
            [[1, 2, 3], [4, 5, 6]],
            index=[datetime(1999, 12, 13), datetime(1999, 12, 14)],
            columns=["c", "d", "c"],
        )
        return df

    @func
    def read_workbook_caller():
        wb = xw.Book.caller()
        return wb.sheets.active["E277"].value == 1


@func
def default_args(x, y="hello", z=20):
    return 2 * x + 3 * len(y) + 7 * z


@func
def variable_args(x, *z):
    return 2 * x + 3 * len(z) + 7 * z[0]


@func
def optional_args(x, y=None):
    if y is None:
        y = 10
    return x * y


@func
def write_none():
    return None


@func
def method_signature_with_less_than_1024_characters(
    very_long_parameter_name_1=None,
    very_long_parameter_name_2=None,
    very_long_parameter_name_3=None,
    very_long_parameter_name_4=None,
    very_long_parameter_name_5=None,
    very_long_parameter_name_6=None,
    very_long_parameter_name_7=None,
    very_long_parameter_name_8=None,
    very_long_parameter_name_9=None,
    very_long_parameter_name_10=None,
    very_long_parameter_name_11=None,
    very_long_parameter_name_12=None,
    very_long_parameter_name_13=None,
    very_long_parameter_name_14=None,
    very_long_parameter_name_15=None,
    very_long_parameter_name_16=None,
    very_long_parameter_name_17=None,
    very_long_parameter_name_18=None,
    very_long_parameter_name_19=None,
    very_long_parameter_name_20=None,
    very_long_parameter_name_21=None,
    very_long_parameter_name_22=None,
    very_long_parameter_name_23=None,
    very_long_parameter_name_24=None,
    very_long_parameter_name_25=None,
    paramet_name_26=None,
):
    return "non splitted signature"


@func
def method_signature_with_more_than_1024_characters(
    very_long_parameter_name_1=None,
    very_long_parameter_name_2=None,
    very_long_parameter_name_3=None,
    very_long_parameter_name_4=None,
    very_long_parameter_name_5=None,
    very_long_parameter_name_6=None,
    very_long_parameter_name_7=None,
    very_long_parameter_name_8=None,
    very_long_parameter_name_9=None,
    very_long_parameter_name_10=None,
    very_long_parameter_name_11=None,
    very_long_parameter_name_12=None,
    very_long_parameter_name_13=None,
    very_long_parameter_name_14=None,
    very_long_parameter_name_15=None,
    very_long_parameter_name_16=None,
    very_long_parameter_name_17=None,
    very_long_parameter_name_18=None,
    very_long_parameter_name_19=None,
    very_long_parameter_name_20=None,
    very_long_parameter_name_21=None,
    very_long_parameter_name_22=None,
    very_long_parameter_name_23=None,
    very_long_parameter_name_24=None,
    very_long_parameter_name_25=None,
    very_long_parameter_name_26=None,
):
    return "splitted signature"


@func
def return_pd_nat():
    return pd.DataFrame(data=[pd.NaT], columns=[1], index=[1])


@func
@arg("df", pd.DataFrame, parse_dates=[0, 2])
def parse_dates_index(df):
    expected = pd.DataFrame(
        [
            [1, dt.datetime(2021, 1, 1, 11, 11, 11), 4],
            [2, dt.datetime(2021, 1, 2, 22, 22, 22), 5],
            [3, dt.datetime(2021, 1, 3), 6],
        ],
        columns=["one", "two", "three"],
        index=[
            dt.datetime(2021, 1, 1, 11, 11, 11),
            dt.datetime(2021, 1, 2, 22, 22, 22),
            dt.datetime(2021, 1, 3),
        ],
    )
    assert_frame_equal(df, expected)
    return True


@func
@arg("df", pd.DataFrame, parse_dates=["ix", "two"])
def parse_dates_names(df):
    expected = pd.DataFrame(
        [
            [1, dt.datetime(2021, 1, 1, 11, 11, 11), 4],
            [2, dt.datetime(2021, 1, 2, 22, 22, 22), 5],
            [3, dt.datetime(2021, 1, 3), 6],
        ],
        columns=["one", "two", "three"],
        index=[
            dt.datetime(2021, 1, 1, 11, 11, 11),
            dt.datetime(2021, 1, 2, 22, 22, 22),
            dt.datetime(2021, 1, 3),
        ],
    )
    expected.index.name = "ix"
    assert_frame_equal(df, expected)
    return True


@func
@arg("df", pd.DataFrame, parse_dates=True)
def parse_dates_true(df):
    expected = pd.DataFrame(
        [[1], [2], [3]],
        columns=["one"],
        index=[
            dt.datetime(2021, 1, 1, 11, 11, 11),
            dt.datetime(2021, 1, 2, 22, 22, 22),
            dt.datetime(2021, 1, 3),
        ],
    )
    assert_frame_equal(df, expected)
    return True


@func
@ret(transpose=True)
def write_error_cells():
    return ["#DIV/0!", "#N/A", "#NAME?", "#NULL!", "#NUM!", "#REF!", "#VALUE!"]


@func
def read_error_cells(errors):
    assert [None] * 7 == errors
    return True


@func
@arg("errors", err_to_str=True)
def read_error_cells_str(errors):
    assert [
        "#DIV/0!",
        "#N/A",
        "#NAME?",
        "#NULL!",
        "#NUM!",
        "#REF!",
        "#VALUE!",
    ] == errors
    return True


@func
@ret(date_format="yyyy-m-d")
def explicit_date_format():
    return dt.datetime(2022, 1, 13)


@func(namespace="subname")
def namespace():
    return True


@func(volatile=True)
def volatile():
    return True



"""
Required Notice: Copyright (C) Zoomer Analytics GmbH.

xlwings PRO is dual-licensed under one of the following licenses:

* PolyForm Noncommercial License 1.0.0 (for noncommercial use):
  https://polyformproject.org/licenses/noncommercial/1.0.0
* xlwings PRO License (for commercial use):
  https://github.com/xlwings/xlwings/blob/main/LICENSE_PRO.txt

Commercial licenses can be purchased at https://www.xlwings.org
"""

import datetime as dt
import numbers
from pathlib import Path

try:
    import numpy as np
except ImportError:
    np = None
try:
    import pandas as pd
except ImportError:
    pd = None

from xlwings import xlwingslib

from .. import NoSuchObjectError, base_classes, utils

MAX_ROWS = 1_048_576
MAX_COLUMNS = 16_384


def _clean_value_data_element(value, datetime_builder, empty_as, number_builder):
    if value is None:
        return empty_as
    elif isinstance(value, dt.datetime) and datetime_builder is not dt.datetime:
        value = datetime_builder(
            month=value.month,
            day=value.day,
            year=value.year,
            hour=value.hour,
            minute=value.minute,
            second=value.second,
            microsecond=value.microsecond,
            tzinfo=None,
        )
    elif number_builder is not None and type(value) == float:
        value = number_builder(value)
    return value


class Engine:
    def __init__(self):
        self.apps = Apps()

    @staticmethod
    def clean_value_data(data, datetime_builder, empty_as, number_builder, err_to_str):
        # err_to_str is handled in raw_values for efficiency
        if empty_as or number_builder or datetime_builder is not dt.datetime:
            return [
                [
                    _clean_value_data_element(
                        c, datetime_builder, empty_as, number_builder
                    )
                    for c in row
                ]
                for row in data
            ]
        else:
            return data

    @staticmethod
    def prepare_xl_data_element(x, options):
        return x

    @property
    def name(self):
        return "calamine"

    @property
    def type(self):
        return "reader"


class Apps(base_classes.Apps):
    def __init__(self):
        self._apps = [App(self)]

    def __iter__(self):
        return iter(self._apps)

    def __len__(self):
        return len(self._apps)

    def __getitem__(self, index):
        return self._apps[index]

    def add(self, **kwargs):
        self._apps.insert(0, App(self, **kwargs))
        return self._apps[0]


class App(base_classes.App):
    _next_pid = -1

    def __init__(self, apps, add_book=True, **kwargs):
        self.apps = apps
        self._pid = App._next_pid
        App._next_pid -= 1
        self._books = Books(self)
        if add_book:
            self._books.add()

    def kill(self):
        self.apps._apps.remove(self)
        self.apps = None

    @property
    def engine(self):
        return engine

    @property
    def books(self):
        return self._books

    @property
    def pid(self):
        return self._pid

    @property
    def visible(self):
        return True

    def activate(self, steal_focus=None):
        pass


class Books(base_classes.Books):
    def __init__(self, app):
        self.app = app
        self.books = []
        self._active = None

    @property
    def active(self):
        return self._active

    def open(self, filename):
        filename = str(Path(filename).resolve())
        sheet_names = xlwingslib.get_sheet_names(filename)
        names = []
        for name, ref in xlwingslib.get_defined_names(filename):
            if ref.split("!")[0].strip("'") in sheet_names:
                names.append(
                    {
                        "name": name,
                        "sheet_index": sheet_names.index(ref.split("!")[0].strip("'")),
                        "address": ref.split("!")[1],
                        "book_scope": True,  # TODO: not provided by calamine
                    }
                )
        book = Book(
            api={
                "sheet_names": sheet_names,
                "names": names,
            },
            books=self,
            path=filename,
        )
        self.books.append(book)
        self._active = book
        return book

    def add(self):
        book = Book(api={"sheet_names": ["Sheet1"]}, books=self, path="dummy")
        self.books.append(book)
        self._active = book
        return book

    def _try_find_book_by_name(self, name):
        for book in self.books:
            if book.name == name or book.fullname == name:
                return book
        return None

    def __len__(self):
        return len(self.books)

    def __iter__(self):
        for book in self.books:
            yield book

    def __call__(self, name_or_index):
        if isinstance(name_or_index, numbers.Number):
            return self.books[name_or_index - 1]
        else:
            book = self._try_find_book_by_name(name_or_index)
            if book is None:
                raise KeyError(name_or_index)
            return book


class Book(base_classes.Book):
    def __init__(self, api, books, path):
        self._api = api
        self.books = books
        self.path = path

    @property
    def api(self):
        return self._api

    @property
    def name(self):
        return Path(self.fullname).name

    @property
    def fullname(self):
        return self.path

    @property
    def names(self):
        return Names(parent=self, api=self.api["names"])

    @property
    def sheets(self):
        return Sheets(book=self)

    @property
    def app(self):
        return self.books.app

    def close(self):
        assert self.api is not None, "Seems this book was already closed."
        self.books.books.remove(self)
        self.books = None
        self._api = None

    def activate(self):
        pass


class Sheets(base_classes.Sheets):
    def __init__(self, book):
        self.book = book

    @property
    def api(self):
        return None

    def __call__(self, name_or_index):
        if isinstance(name_or_index, str):
            sheet_names = self.book.api["sheet_names"]
            if name_or_index not in sheet_names:
                raise NoSuchObjectError(f"Sheet {name_or_index} doesn't exist.")
            else:
                ix = self.book.api["sheet_names"].index(name_or_index) + 1
        else:
            ix = name_or_index

        return Sheet(book=self.book, sheet_index=ix)

    def __len__(self):
        return len(self.book.api["sheet_names"])

    def __iter__(self):
        for ix, sheet in enumerate(self.book.api["sheet_names"]):
            yield Sheet(book=self.book, sheet_index=ix + 1)


class Sheet(base_classes.Sheet):
    def __init__(self, book, sheet_index):
        self._api = {}  # used by e.g., Range.end()
        self._book = book
        self.sheet_index = sheet_index

    @property
    def api(self):
        return self._api

    @property
    def name(self):
        return self.book.api["sheet_names"][self.index - 1]

    @property
    def index(self):
        return self.sheet_index

    @property
    def book(self):
        return self._book

    def range(self, arg1, arg2=None):
        return Range(sheet=self, book=self.book, arg1=arg1, arg2=arg2)

    @property
    def cells(self):
        return Range(
            sheet=self,
            book=self.book,
            arg1=(1, 1),
            arg2=(MAX_ROWS, MAX_COLUMNS),
        )


class Range(base_classes.Range):
    def __init__(self, sheet, book, arg1, arg2=None):
        self.sheet = sheet
        self.book = book
        self.options = None  # Assigned by main.Range to keep API of sheet.range clean

        # Range
        if isinstance(arg1, Range) and isinstance(arg2, Range):
            cell1 = arg1.coords[1], arg1.coords[2]
            cell2 = arg2.coords[1], arg2.coords[2]
            arg1 = min(cell1[0], cell2[0]), min(cell1[1], cell2[1])
            arg2 = max(cell1[0], cell2[0]), max(cell1[1], cell2[1])
        # A1 notation
        if isinstance(arg1, str):
            # A1 notation
            tuple1, tuple2 = utils.a1_to_tuples(arg1)
            if not tuple1:
                # Named range
                for api_name in sheet.book.api["names"]:
                    if (
                        api_name["name"].split("!")[-1] == arg1
                        and api_name["sheet_index"] == sheet.index - 1
                    ):
                        tuple1, tuple2 = utils.a1_to_tuples(api_name["address"])
                if not tuple1:
                    raise NoSuchObjectError(
                        f"The address/named range '{arg1}' doesn't exist."
                    )
            arg1, arg2 = tuple1, tuple2

        # Coordinates
        if len(arg1) == 4:
            row, col, nrows, ncols = arg1
            arg1 = (row, col)
            if nrows > 1 or ncols > 1:
                arg2 = (row + nrows - 1, col + ncols - 1)

        self.arg1 = arg1  # 1-based tuple
        self.arg2 = arg2  # 1-based tuple

    @property
    def api(self):
        return self.raw_value

    @property
    def coords(self):
        return self.sheet.name, self.row, self.column, len(self.api), len(self.api[0])

    @property
    def row(self):
        return self.arg1[0]

    @property
    def column(self):
        return self.arg1[1]

    @property
    def shape(self):
        if self.arg2:
            return self.arg2[0] - self.arg1[0] + 1, self.arg2[1] - self.arg1[1] + 1
        else:
            return 1, 1

    @property
    def raw_value(self):
        err_to_str = self.options.get("err_to_str", False)
        if self.arg2 is None:
            self.arg2 = self.arg1
        if self.arg2[0] == MAX_ROWS and self.arg2[1] == MAX_COLUMNS:
            # Whole sheet via sheet.cells
            if not self.sheet.api.get(f"values_err_to_str_{err_to_str}"):
                values = xlwingslib.get_sheet_values(
                    self.book.fullname, self.sheet.index - 1, err_to_str
                )
                self.sheet._api[f"values_err_to_str_{err_to_str}"] = values
                return values
        else:
            # Specific range
            return xlwingslib.get_range_values(
                self.book.fullname,
                self.sheet.index - 1,
                (self.arg1[0] - 1, self.arg1[1] - 1),
                (self.arg2[0] - 1, self.arg2[1] - 1),
                err_to_str,
            )

    @property
    def address(self):
        nrows, ncols = self.shape
        address = f"${utils.col_name(self.column)}${self.row}"
        if nrows == 1 and ncols == 1:
            return address
        else:
            return (
                f"{address}"
                f":${utils.col_name(self.column + ncols - 1)}${self.row + nrows - 1}"
            )

    @property
    def has_array(self):
        # Not supported, but since this is only used for legacy CSE arrays, probably
        # not much of an issue. Here as there's currently a dependency in expansion.py.
        return None

    def end(self, direction):
        err_to_str = self.options.get("err_to_str", False)
        if not self.sheet.api.get(f"values_err_to_str_{err_to_str}"):
            values = xlwingslib.get_sheet_values(
                self.book.fullname, self.sheet.index - 1, err_to_str
            )
            self.sheet._api[f"values_err_to_str_{err_to_str}"] = values
        else:
            values = self.sheet.api[f"values_err_to_str_{err_to_str}"]
        if direction == "down":
            i = 1
            while True:
                try:
                    if values[self.row - 1 + i][self.column - 1]:
                        i += 1
                    else:
                        break
                except IndexError:
                    break  # outside used range
            nrows = i - 1
            return self.sheet.range((self.row + nrows, self.column))
        if direction == "up":
            i = -1
            while True:
                row_ix = self.row - 1 + i
                if row_ix >= 0 and values[row_ix][self.column - 1]:
                    i -= 1
                else:
                    break
            nrows = i + 1
            return self.sheet.range((self.row + nrows, self.column))
        if direction == "right":
            i = 1
            while True:
                try:
                    if values[self.row - 1][self.column - 1 + i]:
                        i += 1
                    else:
                        break
                except IndexError:
                    break  # outside used range
            ncols = i - 1
            return self.sheet.range((self.row, self.column + ncols))
        if direction == "left":
            i = -1
            while True:
                col_ix = self.column - 1 + i
                if col_ix >= 0 and values[self.row - 1][col_ix]:
                    i -= 1
                else:
                    break
            ncols = i + 1
            return self.sheet.range((self.row, self.column + ncols))

    def __len__(self):
        nrows, ncols = self.shape
        return nrows * ncols

    def __call__(self, arg1, arg2=None):
        if arg2 is None:
            col = (arg1 - 1) % self.shape[1]
            row = int((arg1 - 1 - col) / self.shape[1])
            return self(row + 1, col + 1)
        else:
            return Range(
                sheet=self.sheet,
                book=self.book,
                arg1=(self.row + arg1 - 1, self.column + arg2 - 1),
            )


class Name(base_classes.Name):
    def __init__(self, parent, api):
        self.parent = parent  # only implemented for Book, not Sheet
        self.api = api

    @property
    def name(self):
        return self.api["name"]

    @property
    def refers_to(self):
        sheet_name = self.parent.sheets(self.api["sheet_index"] + 1).name
        sheet_name = f"'{sheet_name}'" if " " in sheet_name else sheet_name
        return f"={sheet_name}!{self.api['address']}"

    @property
    def refers_to_range(self):
        return self.parent.sheets(self.api["sheet_index"] + 1).range(
            self.api["address"]
        )


class Names(base_classes.Names):
    def __init__(self, parent, api):
        self.parent = parent  # only implemented for Book, not Sheet
        self.api = api

    def __call__(self, name_or_index):
        if isinstance(name_or_index, numbers.Number):
            name_or_index -= 1
            if name_or_index > len(self):
                raise KeyError(name_or_index)
            else:
                return Name(self.parent, api=self.api[name_or_index])
        else:
            for ix, i in enumerate(self.api):
                if i["name"] == name_or_index:
                    return Name(self.parent, api=self.api[ix])
            raise KeyError(name_or_index)

    def contains(self, name_or_index):
        if isinstance(name_or_index, numbers.Number):
            return 1 <= name_or_index <= len(self)
        else:
            for i in self.api:
                if i["name"] == name_or_index:
                    return True
            return False

    def __len__(self):
        return len(self.api)


engine = Engine()



"""
Requires the following env var: TEST_XLWINGS_LICENSE_KEY
If you run this on a built/installed package, make sure to cd out of the xlwings source
directory, copy the test folder next to the install xlwings package,then run:

* all tests (this relies on the settings in pytest.ini):

pytest test_e2e.py

* single test:
pytest test_e2e.py::test_name
"""

import os
import shutil
import subprocess
from pathlib import Path
from shlex import split

import pytest

import xlwings as xw

this_dir = Path(__file__).resolve().parent


@pytest.fixture
def app():
    with xw.App(visible=False) as app:
        yield app


@pytest.fixture
def clear_user_config():
    if (Path.home() / ".backup_xlwings").exists():
        shutil.rmtree(Path.home() / ".backup_xlwings")
    if (Path.home() / ".xlwings").exists():
        shutil.copytree(Path.home() / ".xlwings", Path.home() / ".backup_xlwings")
        shutil.rmtree(Path.home() / ".xlwings")
    yield
    if (Path.home() / ".xlwings").exists():
        shutil.rmtree(Path.home() / ".xlwings")
    if (Path.home() / ".backup_xlwings").exists():
        shutil.copytree(Path.home() / ".backup_xlwings", Path.home() / ".xlwings")


@pytest.fixture
def addin(app):
    return app.books.open(Path(xw.__path__[0]) / "addin" / "xlwings.xlam")


@pytest.fixture
def quickstart_book(app, tmpdir):
    os.chdir(tmpdir)
    subprocess.run(split("xlwings quickstart testproject"))
    return app.books.open(Path(tmpdir) / "testproject" / "testproject.xlsm")


def test_config(clear_user_config, app, addin):
    get_config = addin.macro("GetConfig")
    assert get_config("PYTHONPATH") == ""

    # Workbook sheet config
    book = app.books.open(this_dir / "test book.xlsx")
    sheet = book.sheets[0]
    sheet.name = "xlwings.conf"
    sheet["A1"].value = ["PYTHONPATH", "workbook sheet"]

    # Addin sheet config
    addin.sheets[0].name = "xlwings.conf"
    addin.sheets[0]["A1"].value = ["PYTHONPATH", "addin sheet"]

    # Config file workbook directory
    with open(this_dir / "xlwings.conf", "w") as config:
        config.write('"PYTHONPATH","directory config"')

    # Config file user home directory
    os.makedirs(Path.home() / ".xlwings", exist_ok=True)
    with open((Path.home() / ".xlwings" / "xlwings.conf"), "w") as config:
        config.write('"PYTHONPATH","user config"')

    assert get_config("PYTHONPATH") == "workbook sheet"
    sheet.name = "_xlwings.conf"
    assert get_config("PYTHONPATH") == "addin sheet"
    addin.sheets[0].name = "_xlwings.conf"
    assert get_config("PYTHONPATH") == "directory config"
    (this_dir / "xlwings.conf").unlink()
    assert get_config("PYTHONPATH") == "user config"


def test_runpython(addin, quickstart_book):
    sample_call = quickstart_book.macro("Module1.SampleCall")
    sample_call()
    assert quickstart_book.sheets[0]["A1"].value == "Hello xlwings!"
    sample_call()
    assert quickstart_book.sheets[0]["A1"].value == "Bye xlwings!"


def test_runpython_server(addin, quickstart_book):
    sample_call = quickstart_book.macro("Module1.SampleCall")
    quickstart_book.sheets["_xlwings.conf"].name = "xlwings.conf"
    quickstart_book.sheets["xlwings.conf"]["B8"].value = True
    sample_call()
    assert quickstart_book.sheets[0]["A1"].value == "Hello xlwings!"
    sample_call()
    assert quickstart_book.sheets[0]["A1"].value == "Bye xlwings!"


def test_runpython_embedded_code(clear_user_config, addin, quickstart_book):
    os.makedirs(Path.home() / ".xlwings")
    with open((Path.home() / ".xlwings" / "xlwings.conf"), "w") as config:
        config.write(f'"LICENSE_KEY","{os.getenv("TEST_XLWINGS_LICENSE_KEY")}"')
    os.chdir(Path(quickstart_book.fullname).parent)
    subprocess.run(split("xlwings code embed"))
    (Path(quickstart_book.fullname).parent / "testproject.py").unlink()
    sample_call = quickstart_book.macro("Module1.SampleCall")
    sample_call()
    assert quickstart_book.sheets[0]["A1"].value == "Hello xlwings!"
    sample_call()
    assert quickstart_book.sheets[0]["A1"].value == "Bye xlwings!"


def test_udf(clear_user_config, addin, quickstart_book):
    addin.macro("ImportPythonUDFs")()
    quickstart_book.sheets[0]["A1"].value = '=hello("test")'
    assert quickstart_book.sheets[0]["A1"].value == "Hello test!"


def test_udf_embedded_code(clear_user_config, addin, quickstart_book):
    os.makedirs(Path.home() / ".xlwings")
    with open((Path.home() / ".xlwings" / "xlwings.conf"), "w") as config:
        config.write(f'"LICENSE_KEY","{os.getenv("TEST_XLWINGS_LICENSE_KEY")}"')
    os.chdir(Path(quickstart_book.fullname).parent)
    subprocess.run(split("xlwings code embed"))
    (Path(quickstart_book.fullname).parent / "testproject.py").unlink()
    addin.macro("ImportPythonUDFs")()
    quickstart_book.sheets[0]["A1"].value = '=hello("test")'
    assert quickstart_book.sheets[0]["A1"].value == "Hello test!"
    (Path.home() / ".xlwings" / "xlwings.conf").unlink()
    quickstart_book.app.api.CalculateFull()
    assert (
        "xlwings.LicenseError: Embedded code requires a valid LICENSE_KEY."
        in quickstart_book.sheets[0]["A1"].value
    )


def test_can_use_xlwings_without_license_key(clear_user_config, tmp_path):
    import xlwings

    os.chdir(tmp_path)
    subprocess.run(split("xlwings quickstart testproject"))


def test_can_use_xlwings_with_wrong_license_key(clear_user_config, tmp_path):
    os.makedirs(Path.home() / ".xlwings")
    with open((Path.home() / ".xlwings" / "xlwings.conf"), "w") as config:
        config.write(f'"LICENSE_KEY","xxx"')
    import xlwings

    os.chdir(tmp_path)
    subprocess.run(split("xlwings quickstart testproject"))


def test_cant_use_xlwings_pro_without_license_key(clear_user_config):
    with pytest.raises(xw.LicenseError):
        import xlwings.pro


def test_addin_installation(app):
    assert not (Path(app.startup_path) / "xlwings.xlam").exists()
    subprocess.run(split("xlwings addin install"))
    assert (Path(app.startup_path) / "xlwings.xlam").exists()
    subprocess.run(split("xlwings addin remove"))
    assert not (Path(app.startup_path) / "xlwings.xlam").exists()

    # Custom file
    assert not (Path(app.startup_path) / "test book.xlsx").exists()
    os.chdir(this_dir)
    subprocess.run(split('xlwings addin install -f "test book.xlsx"'))
    assert (Path(app.startup_path) / "test book.xlsx").exists()
    subprocess.run(split('xlwings addin remove -f "test book.xlsx"'))
    assert not (Path(app.startup_path) / "test book.xlsx").exists()


def test_update_license_key(clear_user_config):
    subprocess.run(split("xlwings license update -k test_key"))
    with open(Path.home() / ".xlwings" / "xlwings.conf", "r") as f:
        assert f.read() == '"LICENSE_KEY","test_key"\n'


@pytest.mark.skipif(xw.__version__ == "dev", reason="requires a built package")
def test_standalone(clear_user_config, app, tmp_path):
    os.chdir(tmp_path)
    subprocess.run(split("xlwings quickstart testproject --standalone"))
    standalone_book = app.books.open(tmp_path / "testproject" / "testproject.xlsm")
    sample_call = standalone_book.macro("Module1.SampleCall")
    sample_call()
    assert standalone_book.sheets[0]["A1"].value == "Hello xlwings!"
    sample_call()
    assert standalone_book.sheets[0]["A1"].value == "Bye xlwings!"


@pytest.mark.skipif(xw.__version__ == "dev", reason="requires a built package")
def test_runpython_embedded_code_standalone(app, clear_user_config, tmp_path):
    os.chdir(tmp_path)
    subprocess.run(split(f"xlwings quickstart testproject --standalone"))
    quickstart_book = app.books.open(tmp_path / "testproject" / "testproject.xlsm")

    os.makedirs(Path.home() / ".xlwings")
    with open((Path.home() / ".xlwings" / "xlwings.conf"), "w") as config:
        config.write(f'"LICENSE_KEY","{os.getenv("TEST_XLWINGS_LICENSE_KEY")}"')

    os.chdir(tmp_path / "testproject")
    subprocess.run(split("xlwings code embed"))
    (tmp_path / "testproject" / f"testproject.py").unlink()
    sample_call = quickstart_book.macro("Module1.SampleCall")
    sample_call()
    assert quickstart_book.sheets[0]["A1"].value == "Hello xlwings!"
    sample_call()
    assert quickstart_book.sheets[0]["A1"].value == "Bye xlwings!"


@pytest.mark.skipif(xw.__version__ == "dev", reason="requires a built package")
def test_udf_embedded_code_standalone(clear_user_config, app, tmp_path):
    os.chdir(tmp_path)
    subprocess.run(split(f"xlwings quickstart testproject --standalone"))
    quickstart_book = app.books.open(tmp_path / "testproject" / "testproject.xlsm")

    os.makedirs(Path.home() / ".xlwings")
    with open((Path.home() / ".xlwings" / "xlwings.conf"), "w") as config:
        config.write(f'"LICENSE_KEY","{os.getenv("TEST_XLWINGS_LICENSE_KEY")}"')

    os.chdir(tmp_path / "testproject")
    subprocess.run(split("xlwings code embed"))
    (tmp_path / "testproject" / f"testproject.py").unlink()

    quickstart_book.macro("ImportPythonUDFs")()
    quickstart_book.sheets[0]["A1"].value = '=hello("test")'
    assert quickstart_book.sheets[0]["A1"].value == "Hello test!"
    (Path.home() / ".xlwings" / "xlwings.conf").unlink()
    quickstart_book.app.api.CalculateFull()
    assert (
        "xlwings.LicenseError: Embedded code requires a valid LICENSE_KEY."
        in quickstart_book.sheets[0]["A1"].value
    )



"""
Custom Functions (UDFs)
"""

from xlwings import pro


@pro.func
def add(first, second, third=None):
    if third:
        return first + second + third
    else:
        return first + second


@pro.func
async def add2(first, second):
    return first + second



import sys
import unittest

import xlwings as xw

from .common import TestBase


class TestSheets(TestBase):
    def test_active(self):
        self.assertEqual(self.wb2.sheets.active.name, self.wb2.sheets[0].name)

    def test_index(self):
        self.assertEqual(self.wb1.sheets[0].name, self.wb1.sheets(1).name)

    def test_len(self):
        self.assertEqual(len(self.wb1.sheets), 3)

    def del_sheet(self):
        name = self.wb1.sheets[0].name
        del self.wb1.sheets[0]
        self.assertEqual(len(self.wb1.sheets), 2)
        self.assertFalse(self.wb1.sheets[0].name, name)

    def test_iter(self):
        for ix, sht in enumerate(self.wb1.sheets):
            self.assertEqual(self.wb1.sheets[ix].name, sht.name)

    def test_add(self):
        self.wb1.sheets.add()
        self.assertEqual(len(self.wb1.sheets), 4)

    def test_add_before(self):
        new_sheet = self.wb1.sheets.add(before="Sheet1")
        self.assertEqual(self.wb1.sheets[0].name, new_sheet.name)

    def test_add_after(self):
        self.wb1.sheets.add(after=len(self.wb1.sheets))
        self.assertEqual(
            self.wb1.sheets[(len(self.wb1.sheets) - 1)].name,
            self.wb1.sheets.active.name,
        )

        self.wb1.sheets.add(after=1)
        self.assertEqual(self.wb1.sheets[1].name, self.wb1.sheets.active.name)

    def test_add_default(self):
        current_index = self.wb1.sheets.active.index
        self.wb1.sheets.add()
        self.assertEqual(self.wb1.sheets.active.index, current_index)

    def test_add_named(self):
        self.wb1.sheets.add("test", before=1)
        self.assertEqual(self.wb1.sheets[0].name, "test")

    def test_add_name_already_taken(self):
        with self.assertRaises(Exception):
            self.wb1.sheets.add("Sheet1")


class TestSheet(TestBase):
    def test_name(self):
        self.wb1.sheets[0].name = "NewName"
        self.assertEqual(self.wb1.sheets[0].name, "NewName")

    def test_names(self):
        self.wb1.sheets[0].range("A1").name = "test1"
        self.assertEqual(len(self.wb1.sheets[0].names), 0)
        self.wb1.sheets[0].names.add("Sheet1!test2", "Sheet1!B2")
        self.assertEqual(len(self.wb1.sheets[0].names), 1)

    def test_book(self):
        self.assertEqual(self.wb1.sheets[0].book.name, self.wb1.name)

    def test_index(self):
        self.assertEqual(self.wb1.sheets["Sheet1"].index, 1)

    def test_range(self):
        self.wb1.sheets[0].range("A1").value = 123.0
        self.assertEqual(self.wb1.sheets[0].range("A1").value, 123.0)

    def test_cells(self):
        pass  # TODO

    def test_activate(self):
        if sys.platform.startswith("win") and self.app1.version.major > 14:
            # Excel >= 2013 on Win has issues with activating hidden apps correctly
            # over two instances
            with self.assertRaises(Exception):
                self.app1.activate()
        else:
            self.wb2.activate()
            self.wb1.sheets["Sheet2"].activate()
            self.assertEqual(self.wb1.sheets.active.name, "Sheet2")
            self.assertEqual(xw.apps.keys()[0], self.app1.pid)
            self.wb1.sheets[2].activate()
            self.assertEqual(self.wb1.sheets.active.index, 3)
            self.wb1.sheets(1).activate()
            self.assertEqual(self.wb1.sheets.active.index, 1)

    def test_select(self):
        self.wb2.sheets[1].select()
        self.assertEqual(self.wb2.sheets.active, self.wb2.sheets[1])

    def test_clear_content(self):
        self.wb1.sheets["Sheet2"].range("G10").value = 22
        self.wb1.sheets["Sheet2"].clear_contents()
        self.assertEqual(self.wb1.sheets["Sheet2"].range("G10").value, None)

    def test_clear_formats(self):
        self.wb1.sheets["Sheet2"].range("G10").value = 22
        self.wb1.sheets["Sheet2"].range("G10").color = (255, 0, 0)
        self.wb1.sheets["Sheet2"].clear_formats()
        self.assertEqual(self.wb1.sheets["Sheet2"].range("G10").value, 22)
        self.assertEqual(self.wb1.sheets["Sheet2"].range("G10").color, None)

    def test_clear(self):
        self.wb1.sheets["Sheet2"].range("G10").value = 22
        self.wb1.sheets["Sheet2"].range("G10").color = (255, 255, 255)
        self.wb1.sheets["Sheet2"].clear()
        self.assertEqual(self.wb1.sheets["Sheet2"].range("G10").value, None)
        self.assertEqual(self.wb1.sheets["Sheet2"].range("G10").color, None)

    def test_autofit(self):
        sht = self.wb1.sheets["Sheet1"]
        sht.range("A1:D4").value = "test_string"
        sht.range("A1:D4").row_height = 40
        sht.range("A1:D4").column_width = 40
        self.assertEqual(sht.range("A1:D4").row_height, 40)
        self.assertEqual(sht.range("A1:D4").column_width, 40)

        sht.autofit()

        self.assertNotEqual(sht.range("A1:D4").row_height, 40)
        self.assertNotEqual(sht.range("A1:D4").column_width, 40)

        # Just checking if they don't throw an error
        sht.autofit("r")
        sht.autofit("c")
        sht.autofit("rows")
        sht.autofit("columns")

    def test_delete(self):
        self.assertTrue("Sheet1" in [i.name for i in self.wb1.sheets])
        self.wb1.sheets["Sheet1"].delete()
        self.assertFalse("Sheet1" in [i.name for i in self.wb1.sheets])

    def test_used_range(self):
        self.wb1.sheets[0].range("A1:C7").value = 1
        self.assertEqual(
            self.wb1.sheets[0].used_range, self.wb1.sheets[0].range("A1:C7")
        )

    def test_visible(self):
        self.assertTrue(self.wb1.sheets[0].visible)
        self.wb1.sheets[0].visible = False
        self.assertFalse(self.wb1.sheets[0].visible)

    def test_sheet_copy_without_arguments(self):
        original_name = self.wb1.sheets[0].name
        self.wb1.sheets[0]["A1"].value = "xyz"
        self.wb1.sheets[0].copy()
        self.assertEqual(self.wb1.sheets[-1].name, original_name + " (2)")
        self.assertEqual(self.wb1.sheets[-1]["A1"].value, "xyz")

    def test_sheet_copy_with_before_and_after(self):
        with self.assertRaises(AssertionError):
            self.wb1.sheets[0].copy(before=self.wb1.sheets[0], after=self.wb1.sheets[0])

    def test_sheet_copy_before_same_book(self):
        original_name = self.wb1.sheets[0].name
        self.wb1.sheets[0]["A1"].value = "xyz"
        copied_sheet = self.wb1.sheets[0].copy(before=self.wb1.sheets[0])
        self.assertNotEqual(self.wb1.sheets[0].name, original_name)
        self.assertEqual(self.wb1.sheets[0]["A1"].value, "xyz")
        self.assertEqual(copied_sheet.name, self.wb1.sheets[0].name)

    def test_sheet_copy_after_same_book(self):
        original_name = self.wb1.sheets[0].name
        self.wb1.sheets[0]["A1"].value = "xyz"
        self.wb1.sheets[0].copy(after=self.wb1.sheets[0])
        self.assertNotEqual(self.wb1.sheets[1].name, original_name)
        self.assertEqual(self.wb1.sheets[1]["A1"].value, "xyz")

    def test_sheet_copy_before_same_book_new_name(self):
        self.wb1.sheets[0]["A1"].value = "xyz"
        self.wb1.sheets[0].copy(before=self.wb1.sheets[0], name="mycopy")
        self.assertEqual(self.wb1.sheets[0].name, "mycopy")
        self.assertEqual(self.wb1.sheets[0]["A1"].value, "xyz")

    def test_sheet_copy_before_same_book_new_name_already_exists(self):
        self.wb1.sheets[0]["A1"].value = "xyz"
        self.wb1.sheets[0].copy(before=self.wb1.sheets[0], name="mycopy")
        with self.assertRaises(ValueError):
            self.wb1.sheets[0].copy(before=self.wb1.sheets[0], name="mycopy")

    def test_sheet_copy_before_different_book(self):
        self.wb1.sheets[0]["A1"].value = "xyz"
        wb2 = self.wb1.app.books.add()
        self.wb1.sheets[0].copy(before=wb2.sheets[0])
        self.assertEqual(wb2.sheets[0]["A1"].value, self.wb1.sheets[0]["A1"].value)

    def test_sheet_copy_before_different_book_same_name(self):
        mysheet = self.wb1.sheets.add("mysheet")
        mysheet["A1"].value = "xyz"
        wb2 = self.wb1.app.books.add()
        self.wb1.sheets[0].copy(after=wb2.sheets[0], name="mysheet")
        self.assertEqual(wb2.sheets[1]["A1"].value, mysheet["A1"].value)
        self.assertEqual(wb2.sheets[1].name, "mysheet")
        with self.assertRaises(ValueError):
            self.wb1.sheets[0].copy(after=wb2.sheets[0], name="mysheet")


class TestPageSetup(unittest.TestCase):
    def test_print_area(self):
        sheet = xw.Book().sheets[0]
        self.assertIsNone(sheet.page_setup.print_area)
        sheet.page_setup.print_area = "A1:B2"
        self.assertEqual(sheet.page_setup.print_area, "$A$1:$B$2")
        sheet.page_setup.print_area = None
        self.assertIsNone(sheet.page_setup.print_area)
        sheet.book.close()


if __name__ == "__main__":
    unittest.main()



"""
Windows builds currently rely on setup.py instead of pyproject.toml/maturin as long
as the dlls are distributed as data_files
"""
import glob
import os
import re
import sys

from setuptools import find_packages, setup

# long_description: Take from README file
with open(os.path.join(os.path.dirname(__file__), "README.rst")) as f:
    readme = f.read()

# Version Number
with open(os.path.join(os.path.dirname(__file__), "xlwings", "__init__.py")) as f:
    version = re.compile(r'.*__version__ = "(.*?)"', re.S).match(f.read()).group(1)

# Dependencies
data_files = []
install_requires = [
    "pywin32 >= 224;platform_system=='Windows'",
    "psutil >= 2.0.0;platform_system=='Darwin'",
    "appscript >= 1.0.1;platform_system=='Darwin'",
]
if os.environ.get("READTHEDOCS", None) == "True" or os.environ.get(
    "XLWINGS_NO_DEPS"
) in ["1", "True", "true"]:
    # Don't add any further dependencies. Instead of using an env var,
    # you could also run: pip install xlwings --no-deps
    # but when running "pip install -r requirements.txt --no-deps" this would be
    # applied to all packages, which may not be what you want in case the
    # sub-dependencies are not pinned
    pass
elif sys.platform.startswith("win"):
    # This places dlls next to python.exe for standard setup
    # and in the parent folder for virtualenv
    data_files += [("", glob.glob("xlwings??-*.dll"))]
else:
    pass

extras_require = {
    "reports": ["Jinja2", "pdfrw"],
    "all": [
        "black",
        "Jinja2",
        "pandas",
        "matplotlib",
        "plotly",
        "flask",
        "requests",
        "pdfrw",
        "pytest",
    ],
}

if os.getenv("BUILD_RUST", "0") == "1":
    from setuptools_rust import Binding, RustExtension

    rust_extensions = [
        RustExtension(
            "xlwings.xlwingslib",
            binding=Binding.PyO3,
            path="./Cargo.toml",
        )
    ]
else:
    rust_extensions = []

setup(
    name="xlwings",
    version=version,
    rust_extensions=rust_extensions,
    zip_safe=False,  # Rust extensions are not zip safe
    url="https://www.xlwings.org",
    project_urls={
        "Source": "https://github.com/xlwings/xlwings",
        "Documentation": "https://docs.xlwings.org",
    },
    license="BSD 3-clause",
    author="Zoomer Analytics LLC",
    author_email="felix.zumstein@zoomeranalytics.com",
    description="Make Excel fly: Interact with Excel from Python and vice versa.",
    long_description=readme,
    data_files=data_files,
    packages=find_packages(
        exclude=(
            "tests",
            "tests.*",
        )
    ),
    package_data={
        "xlwings": [
            "xlwings.bas",
            "xlwings_custom_addin.bas",
            "*.xlsm",
            "*.xlam",
            "*.applescript",
            "addin/xlwings.xlam",
            "addin/*.cls",
            "addin/WebHelpers.bas",
            "js/xlwings.*",
            "quickstart_fastapi/*.*",
            "html/*.*",
        ],
        "src": [
            "src",
            "Cargo.*",
        ],
    },
    keywords=["xls", "excel", "spreadsheet", "workbook", "vba", "macro"],
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        "console_scripts": ["xlwings=xlwings.cli:main"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Office/Business :: Financial :: Spreadsheet",
        "License :: OSI Approved :: BSD License",
    ],
    python_requires=">=3.7",
)



import os
import sys
import unittest
from datetime import datetime
from pathlib import Path

import xlwings as xw
from xlwings.constants import RgbColor

from .common import TestBase, this_dir


class TestRangeInstantiation(TestBase):
    def test_range1(self):
        r = self.wb1.sheets[0].range("A1")
        self.assertEqual(r.address, "$A$1")

    def test_range2(self):
        r = self.wb1.sheets[0].range("A1:A1")
        self.assertEqual(r.address, "$A$1")

    def test_range3(self):
        r = self.wb1.sheets[0].range("B2:D5")
        self.assertEqual(r.address, "$B$2:$D$5")

    def test_range4(self):
        r = self.wb1.sheets[0].range((1, 1))
        self.assertEqual(r.address, "$A$1")

    def test_range5(self):
        r = self.wb1.sheets[0].range((1, 1), (1, 1))
        self.assertEqual(r.address, "$A$1")

    def test_range6(self):
        r = self.wb1.sheets[0].range((2, 2), (5, 4))
        self.assertEqual(r.address, "$B$2:$D$5")

    def test_range7(self):
        r = self.wb1.sheets[0].range("A1", (2, 2))
        self.assertEqual(r.address, "$A$1:$B$2")

    def test_range8(self):
        r = self.wb1.sheets[0].range((1, 1), "B2")
        self.assertEqual(r.address, "$A$1:$B$2")

    def test_range9(self):
        r = self.wb1.sheets[0].range(
            self.wb1.sheets[0].range("A1"), self.wb1.sheets[0].range("B2")
        )
        self.assertEqual(r.address, "$A$1:$B$2")

    def test_range10(self):
        with self.assertRaises(ValueError):
            self.wb1.sheets[0].range(
                self.wb2.sheets[0].range("A1"), self.wb1.sheets[0].range("B2")
            )

    def test_range11(self):
        with self.assertRaises(ValueError):
            self.wb1.sheets[1].range(
                self.wb1.sheets[0].range("A1"), self.wb1.sheets[0].range("B2")
            )

    def test_range12(self):
        with self.assertRaises(ValueError):
            self.wb1.sheets[0].range(
                self.wb1.sheets[1].range("A1"), self.wb1.sheets[0].range("B2")
            )

    def test_range13(self):
        with self.assertRaises(ValueError):
            self.wb1.sheets[0].range(
                self.wb1.sheets[0].range("A1"), self.wb1.sheets[1].range("B2")
            )

    def test_zero_based_index1(self):
        with self.assertRaises(IndexError):
            self.wb1.sheets[0].range((0, 1)).value = 123

    def test_zero_based_index2(self):
        with self.assertRaises(IndexError):
            self.wb1.sheets[0].range((1, 1), (1, 0)).value

    def test_zero_based_index3(self):
        with self.assertRaises(IndexError):
            xw.Range((1, 0)).value = 123

    def test_zero_based_index4(self):
        with self.assertRaises(IndexError):
            xw.Range((1, 0), (1, 0)).value

    def test_jagged_array(self):
        with self.assertRaises(Exception):
            self.wb1.sheets[0].range("A1").value = [[1], [1, 2]]

        with self.assertRaises(Exception):
            self.wb1.sheets[0].range("A1").value = [[1, 2, 3], [4, 5], [6, 7, 8]]

        with self.assertRaises(Exception):
            self.wb1.sheets[0].range("A1").value = ((1,), (1, 2))

        # the following should not raise an error
        self.wb1.sheets[0].range("A1").value = 1
        self.wb1.sheets[0].range("A1").value = "s"
        self.wb1.sheets[0].range("A1").value = [[1, 2], [1, 2]]
        self.wb1.sheets[0].range("A1").value = [1, 2, 3]
        self.wb1.sheets[0].range("A1").value = [[1, 2, 3]]
        self.wb1.sheets[0].range("A1").value = []


class TestRangeAttributes(TestBase):
    def test_iterator(self):
        self.wb1.sheets[0].range("A20").value = [[1.0, 2.0], [3.0, 4.0]]
        r = self.wb1.sheets[0].range("A20:B21")

        self.assertEqual([c.value for c in r], [1.0, 2.0, 3.0, 4.0])

        # check that reiterating on same range works properly
        self.assertEqual([c.value for c in r], [1.0, 2.0, 3.0, 4.0])

    def test_sheet(self):
        self.assertEqual(
            self.wb1.sheets[1].range("A1").sheet.name, self.wb1.sheets[1].name
        )

    def test_len(self):
        self.assertEqual(len(self.wb1.sheets[0].range("A1:C4")), 12)

    def test_count(self):
        self.assertEqual(
            len(self.wb1.sheets[0].range("A1:C4")),
            self.wb1.sheets[0].range("A1:C4").count,
        )

    def test_row(self):
        self.assertEqual(self.wb1.sheets[0].range("B3:F5").row, 3)

    def test_column(self):
        self.assertEqual(self.wb1.sheets[0].range("B3:F5").column, 2)

    def test_row_count(self):
        self.assertEqual(self.wb1.sheets[0].range("B3:F5").rows.count, 3)

    def test_column_count(self):
        self.assertEqual(self.wb1.sheets[0].range("B3:F5").columns.count, 5)

    def raw_value(self):
        pass  # TODO

    def test_clear_content(self):
        self.wb1.sheets[0].range("G1").value = 22
        self.wb1.sheets[0].range("G1").clear_contents()
        self.assertEqual(self.wb1.sheets[0].range("G1").value, None)

    def test_clear_formats(self):
        self.wb1.sheets[0].range("G1").value = 22
        self.wb1.sheets[0].range("G1").color = (255, 0, 0)
        self.wb1.sheets[0].range("G1").clear_formats()
        self.assertEqual(self.wb1.sheets[0].range("G1").value, 22)
        self.assertEqual(self.wb1.sheets[0].range("G1").color, None)

    def test_clear(self):
        self.wb1.sheets[0].range("G1").value = 22
        self.wb1.sheets[0].range("G1").clear()
        self.assertEqual(self.wb1.sheets[0].range("G1").value, None)

    def test_end(self):
        self.wb1.sheets[0].range("A1:C5").value = 1.0
        self.assertEqual(
            self.wb1.sheets[0].range("A1").end("d"), self.wb1.sheets[0].range("A5")
        )
        self.assertEqual(
            self.wb1.sheets[0].range("A1").end("down"), self.wb1.sheets[0].range("A5")
        )
        self.assertEqual(
            self.wb1.sheets[0].range("C5").end("u"), self.wb1.sheets[0].range("C1")
        )
        self.assertEqual(
            self.wb1.sheets[0].range("C5").end("up"), self.wb1.sheets[0].range("C1")
        )
        self.assertEqual(
            self.wb1.sheets[0].range("A1").end("right"), self.wb1.sheets[0].range("C1")
        )
        self.assertEqual(
            self.wb1.sheets[0].range("A1").end("r"), self.wb1.sheets[0].range("C1")
        )
        self.assertEqual(
            self.wb1.sheets[0].range("C5").end("left"), self.wb1.sheets[0].range("A5")
        )
        self.assertEqual(
            self.wb1.sheets[0].range("C5").end("l"), self.wb1.sheets[0].range("A5")
        )

    def test_formula(self):
        self.wb1.sheets[0].range("A1").formula = "=SUM(A2:A10)"
        self.assertEqual(self.wb1.sheets[0].range("A1").formula, "=SUM(A2:A10)")

    def test_formula2(self):
        self.wb1.sheets[0].range("A1").formula2 = "=UNIQUE(A2:A10)"
        self.assertEqual(self.wb1.sheets[0].range("A1").formula2, "=UNIQUE(A2:A10)")

    def test_formula_array(self):
        self.wb1.sheets[0].range("A1").value = [[1, 4], [2, 5], [3, 6]]
        self.wb1.sheets[0].range("D1").formula_array = "=SUM(A1:A3*B1:B3)"
        self.assertEqual(self.wb1.sheets[0].range("D1").value, 32.0)

    def test_column_width(self):
        self.wb1.sheets[0].range("A1:B2").column_width = 10.0
        result = self.wb1.sheets[0].range("A1").column_width
        self.assertEqual(10.0, result)

        self.wb1.sheets[0].range("A1:B2").value = "ensure cells are used"
        self.wb1.sheets[0].range("B2").column_width = 20.0
        result = self.wb1.sheets[0].range("A1:B2").column_width
        self.assertEqual(None, result)

    def test_row_height(self):
        self.wb1.sheets[0].range("A1:B2").row_height = 15.0
        result = self.wb1.sheets[0].range("A1").row_height
        self.assertEqual(15.0, result)

        self.wb1.sheets[0].range("A1:B2").value = "ensure cells are used"
        self.wb1.sheets[0].range("B2").row_height = 20.0
        result = self.wb1.sheets[0].range("A1:B2").row_height
        self.assertEqual(None, result)

    def test_width(self):
        """test_width: Width depends on default style text size,
        so do not test absolute widths"""
        self.wb1.sheets[0].range("A1:D4").column_width = 10.0
        result_before = self.wb1.sheets[0].range("A1").width
        self.wb1.sheets[0].range("A1:D4").column_width = 12.0
        result_after = self.wb1.sheets[0].range("A1").width
        self.assertTrue(result_after > result_before)

    def test_height(self):
        self.wb1.sheets[0].range("A1:D4").row_height = 60.0
        result = self.wb1.sheets[0].range("A1:D4").height
        self.assertEqual(240.0, result)

    def test_left(self):
        self.assertEqual(self.wb1.sheets[0].range("A1").left, 0.0)
        self.wb1.sheets[0].range("A1").column_width = 20.0
        self.assertEqual(
            self.wb1.sheets[0].range("B1").left, self.wb1.sheets[0].range("A1").width
        )

    def test_top(self):
        self.assertEqual(self.wb1.sheets[0].range("A1").top, 0.0)
        self.wb1.sheets[0].range("A1").row_height = 20.0
        self.assertEqual(
            self.wb1.sheets[0].range("A2").top, self.wb1.sheets[0].range("A1").height
        )

    def test_number_format_cell(self):
        format_string = "mm/dd/yy;@"
        self.wb1.sheets[0].range("A1").number_format = format_string
        result = self.wb1.sheets[0].range("A1").number_format
        self.assertEqual(format_string, result)

    def test_number_format_range(self):
        format_string = "mm/dd/yy;@"
        self.wb1.sheets[0].range("A1:D4").number_format = format_string
        result = self.wb1.sheets[0].range("A1:D4").number_format
        self.assertEqual(format_string, result)

    def test_get_address(self):
        wb1 = self.app1.books.open(os.path.join(this_dir, "test book.xlsx"))

        res = wb1.sheets[0].range((1, 1), (3, 3)).get_address()
        self.assertEqual(res, "$A$1:$C$3")

        res = wb1.sheets[0].range((1, 1), (3, 3)).get_address(False)
        self.assertEqual(res, "$A1:$C3")

        res = wb1.sheets[0].range((1, 1), (3, 3)).get_address(True, False)
        self.assertEqual(res, "A$1:C$3")

        res = wb1.sheets[0].range((1, 1), (3, 3)).get_address(False, False)
        self.assertEqual(res, "A1:C3")

        res = wb1.sheets[0].range((1, 1), (3, 3)).get_address(include_sheetname=True)
        self.assertEqual(res, "'Sheet1'!$A$1:$C$3")

        res = wb1.sheets[1].range((1, 1), (3, 3)).get_address(include_sheetname=True)
        self.assertEqual(res, "'Sheet2'!$A$1:$C$3")

        res = wb1.sheets[0].range((1, 1), (3, 3)).get_address(external=True)
        self.assertEqual(res, "'[test book.xlsx]Sheet1'!$A$1:$C$3")

    def test_address(self):
        self.assertEqual(self.wb1.sheets[0].range("A1:B2").address, "$A$1:$B$2")

    def test_current_region(self):
        values = [[1.0, 2.0], [3.0, 4.0]]
        self.wb1.sheets[0].range("A20").value = values
        self.assertEqual(self.wb1.sheets[0].range("B21").current_region.value, values)

    def test_autofit_range(self):
        self.wb1.sheets[0].range("A1:D4").value = "test_string"

        self.wb1.sheets[0].range("A1:D4").row_height = 40
        self.wb1.sheets[0].range("A1:D4").column_width = 40
        self.assertEqual(40, self.wb1.sheets[0].range("A1:D4").row_height)
        self.assertEqual(40, self.wb1.sheets[0].range("A1:D4").column_width)
        self.wb1.sheets[0].range("A1:D4").autofit()
        self.assertTrue(40 != self.wb1.sheets[0].range("A1:D4").column_width)
        self.assertTrue(40 != self.wb1.sheets[0].range("A1:D4").row_height)

        self.wb1.sheets[0].range("A1:D4").row_height = 40
        self.assertEqual(40, self.wb1.sheets[0].range("A1:D4").row_height)
        self.wb1.sheets[0].range("A1:D4").rows.autofit()
        self.assertTrue(40 != self.wb1.sheets[0].range("A1:D4").row_height)

        self.wb1.sheets[0].range("A1:D4").column_width = 40
        self.assertEqual(40, self.wb1.sheets[0].range("A1:D4").column_width)
        self.wb1.sheets[0].range("A1:D4").columns.autofit()
        self.assertTrue(40 != self.wb1.sheets[0].range("A1:D4").column_width)

        self.wb1.sheets[0].range("A1:D4").rows.autofit()
        self.wb1.sheets[0].range("A1:D4").columns.autofit()

    def test_autofit_col(self):
        self.wb1.sheets[0].range("A1:D4").value = "test_string"
        self.wb1.sheets[0].range("A:D").column_width = 40
        self.assertEqual(40, self.wb1.sheets[0].range("A:D").column_width)
        self.wb1.sheets[0].range("A:D").autofit()
        self.assertTrue(40 != self.wb1.sheets[0].range("A:D").column_width)

        # Just checking if they don't throw an error
        self.wb1.sheets[0].range("A:D").rows.autofit()
        self.wb1.sheets[0].range("A:D").columns.autofit()

    def test_autofit_row(self):
        self.wb1.sheets[0].range("A1:D4").value = "test_string"
        self.wb1.sheets[0].range("1:10").row_height = 40
        self.assertEqual(40, self.wb1.sheets[0].range("1:10").row_height)
        self.wb1.sheets[0].range("1:10").autofit()
        self.assertTrue(40 != self.wb1.sheets[0].range("1:10").row_height)

        # Just checking if they don't throw an error
        self.wb1.sheets[0].range("1:1000000").rows.autofit()
        self.wb1.sheets[0].range("1:1000000").columns.autofit()

    def test_color(self):
        rgb = (30, 100, 200)

        self.wb1.sheets[0].range("A1").color = rgb
        self.assertEqual(rgb, self.wb1.sheets[0].range("A1").color)

        self.wb1.sheets[0].range("A2").color = RgbColor.rgbAqua
        self.assertEqual((0, 255, 255), self.wb1.sheets[0].range("A2").color)

        self.wb1.sheets[0].range("A2").color = None
        self.assertEqual(self.wb1.sheets[0].range("A2").color, None)

        self.wb1.sheets[0].range("A1:D4").color = rgb
        self.assertEqual(rgb, self.wb1.sheets[0].range("A1:D4").color)

    def test_len_rows(self):
        self.assertEqual(len(self.wb1.sheets[0].range("A1:C4").rows), 4)

    def test_count_rows(self):
        self.assertEqual(
            len(self.wb1.sheets[0].range("A1:C4").rows),
            self.wb1.sheets[0].range("A1:C4").rows.count,
        )

    def test_len_cols(self):
        self.assertEqual(len(self.wb1.sheets[0].range("A1:C4").columns), 3)

    def test_count_cols(self):
        self.assertEqual(
            len(self.wb1.sheets[0].range("A1:C4").columns),
            self.wb1.sheets[0].range("A1:C4").columns.count,
        )

    def test_shape(self):
        self.assertEqual(self.wb1.sheets[0].range("A1:C4").shape, (4, 3))

    def test_size(self):
        self.assertEqual(self.wb1.sheets[0].range("A1:C4").size, 12)

    def test_table(self):
        data = [
            [1, 2.222, 3.333],
            ["Test1", None, "éöà"],
            [datetime(1962, 11, 3), datetime(2020, 12, 31, 12, 12, 20), 9.999],
        ]
        self.wb1.sheets[0].range("A1").value = data
        if sys.platform.startswith("win") and self.wb1.app.version == "14.0":
            self.wb1.sheets[0].range(
                "A3:B3"
            ).number_format = "dd/mm/yyyy"  # Hack for Excel 2010 bug, see GH #43
        cells = self.wb1.sheets[0].range("A1").expand("table").value
        self.assertEqual(cells, data)

    def test_vertical(self):
        data = [
            [1, 2.222, 3.333],
            ["Test1", None, "éöà"],
            [datetime(1962, 11, 3), datetime(2020, 12, 31, 12, 12, 20), 9.999],
        ]
        self.wb1.sheets[0].range("A10").value = data
        if sys.platform.startswith("win") and self.wb1.app.version == "14.0":
            self.wb1.sheets[0].range(
                "A12:B12"
            ).number_format = "dd/mm/yyyy"  # Hack for Excel 2010 bug, see GH #43
        cells = self.wb1.sheets[0].range("A10").expand("vertical").value
        self.assertEqual(cells, [row[0] for row in data])

        cells = self.wb1.sheets[0].range("A10").expand("d").value
        self.assertEqual(cells, [row[0] for row in data])

        cells = self.wb1.sheets[0].range("A10").expand("down").value
        self.assertEqual(cells, [row[0] for row in data])

    def test_horizontal(self):
        data = [
            [1, 2.222, 3.333],
            ["Test1", None, "éöà"],
            [datetime(1962, 11, 3), datetime(2020, 12, 31, 12, 12, 20), 9.999],
        ]
        self.wb1.sheets[0].range("A20").value = data
        cells = self.wb1.sheets[0].range("A20").expand("horizontal").value
        self.assertEqual(cells, data[0])

        cells = self.wb1.sheets[0].range("A20").expand("r").value
        self.assertEqual(cells, data[0])

        cells = self.wb1.sheets[0].range("A20").expand("right").value
        self.assertEqual(cells, data[0])

    def test_hyperlink(self):
        address = "www.xlwings.org"
        # Naked address
        self.wb1.sheets[0].range("A1").add_hyperlink(address)
        self.assertEqual(self.wb1.sheets[0].range("A1").value, address)
        hyperlink = self.wb1.sheets[0].range("A1").hyperlink
        if not hyperlink.endswith("/"):
            hyperlink += "/"
        self.assertEqual(hyperlink, "http://" + address + "/")

        # Address + FriendlyName
        self.wb1.sheets[0].range("A2").add_hyperlink(address, "test_link")
        self.assertEqual(self.wb1.sheets[0].range("A2").value, "test_link")
        hyperlink = self.wb1.sheets[0].range("A2").hyperlink
        if not hyperlink.endswith("/"):
            hyperlink += "/"
        self.assertEqual(hyperlink, "http://" + address + "/")

    def test_hyperlink_formula(self):
        self.wb1.sheets[0].range(
            "B10"
        ).formula = '=HYPERLINK("http://xlwings.org", "xlwings")'
        self.assertEqual(
            self.wb1.sheets[0].range("B10").hyperlink, "http://xlwings.org"
        )

    def test_insert_cell(self):
        self.wb1.sheets[0].range("A1:C1").value = "test"
        self.wb1.sheets[0].range("A1").insert()
        self.assertIsNone(self.wb1.sheets[0].range("A1").value)
        self.assertEqual(self.wb1.sheets[0].range("A2").value, "test")

    def test_insert_row(self):
        self.wb1.sheets[0].range("A1:C1").value = "test"
        self.wb1.sheets[0].range("1:1").insert()
        self.assertEqual(self.wb1.sheets[0].range("A1:C1").value, [None, None, None])
        self.assertEqual(
            self.wb1.sheets[0].range("A2:C2").value, ["test", "test", "test"]
        )

    def test_insert_column(self):
        self.wb1.sheets[0].range("A1:A3").value = "test"
        self.wb1.sheets[0].range("A:A").insert()
        self.assertEqual(self.wb1.sheets[0].range("A1:A3").value, [None, None, None])
        self.assertEqual(
            self.wb1.sheets[0].range("B1:B3").value, ["test", "test", "test"]
        )

    def test_insert_cell_shift_down(self):
        self.wb1.sheets[0].range("A1:C1").value = "test"
        self.wb1.sheets[0].range("A1").insert(shift="down")
        self.assertIsNone(self.wb1.sheets[0].range("A1").value)
        self.assertEqual(self.wb1.sheets[0].range("A2").value, "test")

    def test_insert_cell_shift_right(self):
        self.wb1.sheets[0].range("A1:C1").value = "test"
        self.wb1.sheets[0].range("A1").insert(shift="right")
        self.assertIsNone(self.wb1.sheets[0].range("A1").value)
        self.assertEqual(
            self.wb1.sheets[0].range("B1:D1").value, ["test", "test", "test"]
        )

    def test_delete_cell(self):
        self.wb1.sheets[0].range("A1").value = ["one", "two", "three"]
        self.wb1.sheets[0].range("A1").delete()
        self.assertIsNone(self.wb1.sheets[0].range("C1").value)
        self.assertEqual(self.wb1.sheets[0].range("A1").value, "two")

    def test_delete_row(self):
        self.wb1.sheets[0].range("A1:C1").value = "one"
        self.wb1.sheets[0].range("A2:C2").value = "two"
        self.wb1.sheets[0].range("1:1").delete()
        self.assertEqual(self.wb1.sheets[0].range("A1:C1").value, ["two", "two", "two"])
        self.assertEqual(self.wb1.sheets[0].range("A2:C2").value, [None, None, None])

    def test_delete_column(self):
        self.wb1.sheets[0].range("A1:A1").value = "one"
        self.wb1.sheets[0].range("B1:B2").value = "two"
        self.wb1.sheets[0].range("C1:C2").value = "two"
        self.wb1.sheets[0].range("A:A").delete()
        self.assertEqual(self.wb1.sheets[0].range("C1:C2").value, [None, None])
        self.assertEqual(self.wb1.sheets[0].range("A1:A2").value, ["two", "two"])

    def test_delete_cell_shift_up(self):
        self.wb1.sheets[0].range("A1").value = ["one", "two", "three"]
        self.wb1.sheets[0].range("A1").delete("up")
        self.assertIsNone(self.wb1.sheets[0].range("A1").value)
        self.assertEqual(self.wb1.sheets[0].range("B1:C1").value, ["two", "three"])

    def test_delete_cell_shift_left(self):
        self.wb1.sheets[0].range("A1").value = ["one", "two", "three"]
        self.wb1.sheets[0].range("A1").delete("left")
        self.assertIsNone(self.wb1.sheets[0].range("C1").value)
        self.assertEqual(self.wb1.sheets[0].range("A1").value, "two")

    def test_copy_destination(self):
        sheet = self.wb1.sheets[0]
        sheet.range("A1:B1").value = "test"
        sheet.range("A1:B1").copy(destination=sheet.range("A2"))
        self.assertEqual(sheet.range("A1:B1").value, sheet.range("A2:B2").value)

    def test_copy_clipboard(self):
        sheet = self.wb1.sheets[0]
        sheet.range("A1:B1").value = "test"
        sheet.range("A1:B1").copy()

    def test_paste(self):
        sheet = self.wb1.sheets[0]
        sheet.range("A1:B1").value = "test"
        sheet.range("A1:B1").color = (34, 34, 34)
        sheet.range("A1:B1").copy()
        sheet.range("A2").paste()
        self.assertEqual(sheet["A1:B1"].value, sheet["A2:B2"].value)
        self.assertEqual(sheet["A1:B1"].color, sheet["A2:B2"].color)

    def test_paste_values(self):
        sheet = self.wb1.sheets[0]
        sheet.range("A1:B1").value = "test"
        sheet.range("A1:B1").color = (34, 34, 34)
        sheet.range("A1:B1").copy()
        sheet.range("A2").paste(paste="values")
        self.assertEqual(sheet["A1:B1"].value, sheet["A2:B2"].value)
        self.assertNotEqual(sheet["A1:B1"].color, sheet["A2:B2"].color)

    def test_resize(self):
        r = self.wb1.sheets[0].range("A1").resize(4, 5)
        self.assertEqual(r.address, "$A$1:$E$4")

        r = self.wb1.sheets[0].range("A1").resize(row_size=4)
        self.assertEqual(r.address, "$A$1:$A$4")

        r = self.wb1.sheets[0].range("A1:B4").resize(column_size=5)
        self.assertEqual(r.address, "$A$1:$E$4")

        r = self.wb1.sheets[0].range("A1:B4").resize(row_size=5)
        self.assertEqual(r.address, "$A$1:$B$5")

        r = self.wb1.sheets[0].range("A1:B4").resize()
        self.assertEqual(r.address, "$A$1:$B$4")

        r = self.wb1.sheets[0].range("A1:C5").resize(row_size=1)
        self.assertEqual(r.address, "$A$1:$C$1")

        with self.assertRaises(AssertionError):
            self.wb1.sheets[0].range("A1:B4").resize(row_size=0)

        with self.assertRaises(AssertionError):
            self.wb1.sheets[0].range("A1:B4").resize(column_size=0)

    def test_offset(self):
        o = self.wb1.sheets[0].range("A1:B3").offset(3, 4)
        self.assertEqual(o.address, "$E$4:$F$6")

        o = self.wb1.sheets[0].range("A1:B3").offset(row_offset=3)
        self.assertEqual(o.address, "$A$4:$B$6")

        o = self.wb1.sheets[0].range("A1:B3").offset(column_offset=4)
        self.assertEqual(o.address, "$E$1:$F$3")

    def test_last_cell(self):
        self.assertEqual(self.wb1.sheets[0].range("B3:F5").last_cell.row, 5)
        self.assertEqual(self.wb1.sheets[0].range("B3:F5").last_cell.column, 6)

    def test_select(self):
        self.wb2.sheets[0].range("C10").select()
        self.assertEqual(
            self.app2.selection.address, self.wb2.sheets[0].range("C10").address
        )

    def test_wrap_text(self):
        self.assertFalse(self.wb1.sheets[0]["A1"].wrap_text)
        self.wb1.sheets[0]["A1"].wrap_text = True
        self.assertTrue(self.wb1.sheets[0]["A1"].wrap_text)


class TestRangeIndexing(TestBase):
    # 2d Range
    def test_index1(self):
        r = self.wb1.sheets[0].range("A1:B2")

        self.assertEqual(r[0].address, "$A$1")
        self.assertEqual(r(1).address, "$A$1")

        self.assertEqual(r[0, 0].address, "$A$1")
        self.assertEqual(r(1, 1).address, "$A$1")

    def test_index2(self):
        r = self.wb1.sheets[0].range("A1:B2")

        self.assertEqual(r[1].address, "$B$1")
        self.assertEqual(r(2).address, "$B$1")

        self.assertEqual(r[0, 1].address, "$B$1")
        self.assertEqual(r(1, 2).address, "$B$1")

    def test_index3(self):
        with self.assertRaises(IndexError):
            r = self.wb1.sheets[0].range("A1:B2")
            r[4].address

    def test_index4(self):
        r = self.wb1.sheets[0].range("A1:B2")
        self.assertEqual(r(5).address, "$A$3")

    def test_index5(self):
        with self.assertRaises(IndexError):
            r = self.wb1.sheets[0].range("A1:B2")
            r[0, 4].address

    def test_index6(self):
        r = self.wb1.sheets[0].range("A1:B2")
        self.assertEqual(r(1, 5).address, "$E$1")

    # Row
    def test_index1row(self):
        r = self.wb1.sheets[0].range("A1:D1")

        self.assertEqual(r[0].address, "$A$1")
        self.assertEqual(r(1).address, "$A$1")

        self.assertEqual(r[0, 0].address, "$A$1")
        self.assertEqual(r(1, 1).address, "$A$1")

    def test_index2row(self):
        r = self.wb1.sheets[0].range("A1:D1")

        self.assertEqual(r[1].address, "$B$1")
        self.assertEqual(r(2).address, "$B$1")

        self.assertEqual(r[0, 1].address, "$B$1")
        self.assertEqual(r(1, 2).address, "$B$1")

    def test_index3row(self):
        with self.assertRaises(IndexError):
            r = self.wb1.sheets[0].range("A1:D1")
            r[4].address

    def test_index4row(self):
        r = self.wb1.sheets[0].range("A1:D1")
        self.assertEqual(r(5).address, "$A$2")

    def test_index5row(self):
        with self.assertRaises(IndexError):
            r = self.wb1.sheets[0].range("A1:D1")
            r[0, 4].address

    def test_index6row(self):
        r = self.wb1.sheets[0].range("A1:D1")
        self.assertEqual(r(1, 5).address, "$E$1")

    # Column
    def test_index1col(self):
        r = self.wb1.sheets[0].range("A1:A4")

        self.assertEqual(r[0].address, "$A$1")
        self.assertEqual(r(1).address, "$A$1")

        self.assertEqual(r[0, 0].address, "$A$1")
        self.assertEqual(r(1, 1).address, "$A$1")

    def test_index2col(self):
        r = self.wb1.sheets[0].range("A1:A4")

        self.assertEqual(r[1].address, "$A$2")
        self.assertEqual(r(2).address, "$A$2")

        self.assertEqual(r[1, 0].address, "$A$2")
        self.assertEqual(r(2, 1).address, "$A$2")

    def test_index3col(self):
        with self.assertRaises(IndexError):
            r = self.wb1.sheets[0].range("A1:A4")
            r[4].address

    def test_index4col(self):
        r = self.wb1.sheets[0].range("A1:A4")
        self.assertEqual(r(5).address, "$A$5")

    def test_index5col(self):
        with self.assertRaises(IndexError):
            r = self.wb1.sheets[0].range("A1:A4")
            r[4, 0].address

    def test_index6col(self):
        r = self.wb1.sheets[0].range("A1:A4")
        self.assertEqual(r(5, 1).address, "$A$5")


class TestRangeSlicing(TestBase):
    # 2d Range
    def test_slice1(self):
        r = self.wb1.sheets[0].range("B2:D4")
        self.assertEqual(r[0:, 1:].address, "$C$2:$D$4")

    def test_slice2(self):
        r = self.wb1.sheets[0].range("B2:D4")
        self.assertEqual(r[1:2, 1:2].address, "$C$3")

    def test_slice3(self):
        r = self.wb1.sheets[0].range("B2:D4")
        self.assertEqual(r[:1, :2].address, "$B$2:$C$2")

    def test_slice4(self):
        r = self.wb1.sheets[0].range("B2:D4")
        self.assertEqual(r[:, :].address, "$B$2:$D$4")

    # Row
    def test_slice1row(self):
        r = self.wb1.sheets[0].range("B2:D2")
        self.assertEqual(r[1:].address, "$C$2:$D$2")

    def test_slice2row(self):
        r = self.wb1.sheets[0].range("B2:D2")
        self.assertEqual(r[1:2].address, "$C$2")

    def test_slice3row(self):
        r = self.wb1.sheets[0].range("B2:D2")
        self.assertEqual(r[:2].address, "$B$2:$C$2")

    def test_slice4row(self):
        r = self.wb1.sheets[0].range("B2:D2")
        self.assertEqual(r[:].address, "$B$2:$D$2")

    # Column
    def test_slice1col(self):
        r = self.wb1.sheets[0].range("B2:B4")
        self.assertEqual(r[1:].address, "$B$3:$B$4")

    def test_slice2col(self):
        r = self.wb1.sheets[0].range("B2:B4")
        self.assertEqual(r[1:2].address, "$B$3")

    def test_slice3col(self):
        r = self.wb1.sheets[0].range("B2:B4")
        self.assertEqual(r[:2].address, "$B$2:$B$3")

    def test_slice4col(self):
        r = self.wb1.sheets[0].range("B2:B4")
        self.assertEqual(r[:].address, "$B$2:$B$4")


class TestRangeShortcut(TestBase):
    def test_shortcut1(self):
        self.assertEqual(self.wb1.sheets[0]["A1"], self.wb1.sheets[0].range("A1"))

    def test_shortcut2(self):
        self.assertEqual(self.wb1.sheets[0]["A1:B5"], self.wb1.sheets[0].range("A1:B5"))

    def test_shortcut3(self):
        self.assertEqual(self.wb1.sheets[0][0, 1], self.wb1.sheets[0].range("B1"))

    def test_shortcut4(self):
        self.assertEqual(self.wb1.sheets[0][:5, :5], self.wb1.sheets[0].range("A1:E5"))

    def test_shortcut5(self):
        with self.assertRaises(TypeError):
            self.wb1.sheets[0]["A1", "B5"]

    def test_shortcut6(self):
        with self.assertRaises(TypeError):
            self.wb1.sheets[0][self.wb1.sheets[0]["A1"], "B5"]

    def test_shortcut7(self):
        with self.assertRaises(TypeError):
            self.wb1.sheets[0]["A1", self.wb1.sheets[0]["B5"]]


class TestRangeExpansion(TestBase):
    def test_table(self):
        sht = self.wb1.sheets[0]
        rng = sht[0, 0]

        rng.value = [["a"] * 5] * 5

        self.assertEqual(rng.options(expand="table").value, [["a"] * 5] * 5)

    def test_vertical(self):
        sht = self.wb1.sheets[0]
        rng = sht[0, 0:3]

        sht[0, 0].value = [["a"] * 3] * 5

        self.assertEqual(rng.options(expand="down").value, [["a"] * 3] * 5)

    def test_horizontal(self):
        sht = self.wb1.sheets[0]
        rng = sht[0:5, 0]

        sht[0, 0].value = [["a"] * 3] * 5

        self.assertEqual(rng.options(expand="right").value, [["a"] * 3] * 5)

    def test_copy_picture(self):
        rng = self.wb1.sheets[0]["A1:B2"]
        rng.copy_picture()
        rng.copy_picture("screen", "picture")
        rng.copy_picture("printer", "bitmap")


class TestCellErrors(TestBase):
    def test_cell_errors_default(self):
        wb = xw.Book(Path(this_dir) / "cell_errors.xlsx")
        sheet = wb.sheets[0]

        for i in range(1, 8):
            self.assertIsNone(sheet.range((i, 1)).value)
        wb.close()

    def test_cell_errors_str(self):
        wb = xw.Book(Path(this_dir) / "cell_errors.xlsx")
        sheet = wb.sheets[0]
        # Single cells, since macOS has massive issues with ranges that contain cell
        # errors, see #1028 and #1924
        expected = ["#DIV/0!", "#N/A", "#NAME?", "#NULL!", "#NUM!", "#REF!", "#VALUE!"]
        for i in range(1, 8):
            self.assertEqual(
                sheet.range((i, 1)).options(err_to_str=True).value, expected[i - 1]
            )
        wb.close()


class TestMerging(TestBase):
    def test_merge(self):
        sheet = self.wb1.sheets[0]
        self.assertEqual(sheet["A1"].merge_area, sheet["A1"])
        self.assertEqual(sheet["A1"].merge_cells, False)
        sheet["A1:A2"].merge()
        self.assertEqual(sheet["A1"].merge_area, sheet["A1:A2"])
        self.assertEqual(sheet["A1"].merge_cells, True)
        sheet["A1:B2"].merge()
        self.assertEqual(sheet["A1"].merge_area, sheet["A1:B2"])
        sheet["A1:B2"].unmerge()
        self.assertEqual(sheet["A1"].merge_area, sheet["A1"])
        sheet["A1:B2"].merge(True)
        self.assertEqual(sheet["A1"].merge_area, sheet["A1:B1"])


class TestNotes(unittest.TestCase):
    def test_note(self):
        sheet = xw.Book(Path(this_dir) / "reports" / "template_one_frame.xlsx").sheets[
            0
        ]
        self.assertEqual(sheet["A1"].note.text, "<frame>")
        self.assertIsNone(sheet["A2"].note)
        sheet["A1"].note.delete()
        self.assertIsNone(sheet["A1"].note)
        sheet.book.close()


if __name__ == "__main__":
    unittest.main()
